﻿using System.Collections.Generic;
using Volvo.CustomerMaster.Infrastructure.Common.Utilities.Entities;
using Volvo.CustomerMaster.Infrastructure.Common.Utilities.Exceptions;
using Microsoft.Practices.EnterpriseLibrary.Security;

namespace Volvo.CustomerMaster.Infrastructure.Common.Utilities.Security
{
    public class AuthorizationHelper
    {
        /// <summary>
        /// Ensures that the user is authorized to use an entity or some other business object 
        /// implementing the IAuthorizationAware interface
        /// </summary>
        /// <param name="user"></param>
        /// <param name="entity"></param>
        public static void EnsureCredentials(GenericUser user, IAuthorizationAware entity)
        {
            if (user == null || (!(entity.IsAuthorized(user))))
            {
                throw new NotAuthorizedException("User has invalid credentials for managing the entity");
            }
        }

        /// <summary>
        /// Authorize the user roles against selected rules using the security application block
        /// </summary>
        /// <param name="user">user with roles</param>
        /// <param name="rules">rules that is tested against the roles</param>
        /// <returns></returns>
        public static bool Authorize(GenericUser user, IList<string> rules)
        {
            return Authorize(user, rules, null);
        }

        /// <summary>
        /// Authorize the user roles against selected rules using the security application block
        /// </summary>
        /// <param name="user">user with roles</param>
        /// <param name="rules">rules that is tested against the roles</param>
        /// <param name="authorizationProvider">set in security block</param>
        /// <returns></returns>
        public static bool Authorize(GenericUser user, IList<string> rules, string authorizationProvider)
        {
            IAuthorizationProvider auth;

            // Get authorization provider from Entlib security block
            if (authorizationProvider != null)
            {
                auth = AuthorizationFactory.GetAuthorizationProvider(authorizationProvider);
            }
            else
            {
                auth = AuthorizationFactory.GetAuthorizationProvider();
            }

            foreach (string rule in rules)
            {
                // Authorize user roles agains the rule. The combination is defined in the Entlib.config
                if (!auth.Authorize(user.Principal, rule))
                {
                    return false;
                }
            }
            return true;
        }
    }
}
