﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Linq.Expressions;
using System.Web;
using Telerik.Web.Mvc;
using Telerik.Web.Mvc.Infrastructure;
using Volvo.CustomerMaster.Infrastructure.Common.Utilities.Entities;
using Volvo.CustomerMaster.Infrastructure.Common.Utilities.Specifications;
using Entities = Volvo.CustomerMaster.Web.Models.Entities;


namespace Volvo.CustomerMaster.Web.Common.Helpers
{
    public static class TelerikHelper
    {
        /// <summary>
        /// Translates Telerik filter into FilterSpecification
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="filterDescriptors">The filter descriptors.</param>
        /// <returns></returns>
        public static Specification<T> TranslateFilter<T>(IList<IFilterDescriptor> filterDescriptors)
        {
            return new Specification<T>(ExpressionBuilder.Expression<T>(filterDescriptors));
        }

        /// <summary>
        /// Builds the Telerik group.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="TEntity">The type of the entity.</typeparam>
        /// <param name="groupSelector">The group selector.</param>
        /// <param name="selectorBuilder">The selector builder.</param>
        /// <returns></returns>
        private static Expression<Func<IEnumerable<TEntity>, IEnumerable<AggregateFunctionsGroup>>> BuildGroup<T, TEntity>(Func<TEntity, T> groupSelector, Expression<Func<IEnumerable<TEntity>, IEnumerable<AggregateFunctionsGroup>>> selectorBuilder)
        {
            var tempSelector = selectorBuilder;

            return g => g.GroupBy(groupSelector)
                         .Select(c => new AggregateFunctionsGroup
                         {
                             Key = c.Key,
                             HasSubgroups = true,
                             Items = tempSelector.Compile().Invoke(c).ToList()
                         });
        }

        /// <summary>
        /// Builds the Telerik inner group.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="TEntity">The type of the entity.</typeparam>
        /// <param name="group">The group.</param>
        /// <param name="groupSelector">The group selector.</param>
        /// <returns></returns>
        private static IEnumerable<AggregateFunctionsGroup> BuildInnerGroup<T, TEntity>(IEnumerable<TEntity> group, Func<TEntity, T> groupSelector)
        {
            return group.GroupBy(groupSelector)
                    .Select(i => new AggregateFunctionsGroup
                    {
                        Key = i.Key,
                        Items = i.ToList()
                    });
        }

        /// <summary>
        /// Translates grouping Orders into GroupSpecification with Telerik's AggregateFunctionsGroup.
        /// </summary>
        /// <param name="groupDescriptors">The group descriptors.</param>
        /// <returns></returns>
        public static GroupSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer, AggregateFunctionsGroup> TranslateCustomerGroup(IList<GroupDescriptor> groupDescriptors)
        {
            Expression<Func<IEnumerable<Volvo.CustomerMaster.Customer.Model.Entities.Customer>, IEnumerable<AggregateFunctionsGroup>>> selector = null;
            foreach (var group in groupDescriptors.Reverse())
            {
                if (selector == null)
                {
                    switch (group.Member)
                    {
                        case "Number":
                            selector = orders => BuildInnerGroup(orders, o => o.Number);
                            break;
                      
                        case "Name":
                            selector = orders => BuildInnerGroup(orders, o => o.Name);
                            break;

                        case "Address":
                            selector = orders => BuildInnerGroup(orders, o => o.Address );
                            break; 
                      
                    }
                }
                else
                {
                    switch (group.Member)
                    {
                        case "Number":
                            selector = BuildGroup(o => o.Number, selector);
                            break;

                        case "Name":
                            selector = BuildGroup(o => o.Name, selector);
                            break;

                        case "Address":
                            selector = BuildGroup(o => o.Address , selector);
                            break;
                    }
                }
            }
            return (selector == null) ? null : new GroupSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer, AggregateFunctionsGroup>() { Predicate = selector };
        }

        /// <summary>
        /// Translates sort Orders into SortSpecification of Orders.
        /// </summary>
        /// <param name="groupDescriptors">The group descriptors.</param>
        /// <param name="sortDescriptors">The sort descriptors.</param>
        /// <returns></returns>
        public static IEnumerable<SortSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer>> TranslateCustomerSort(IList<GroupDescriptor> groupDescriptors, IList<SortDescriptor> sortDescriptors)
        {
            var definitions = new List<SortSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer>>();

            if (groupDescriptors.Any())
            {
                foreach (var groupDescriptor in groupDescriptors.Reverse())
                {
                    definitions.Add(CreateCustomerSortExpression(groupDescriptor.SortDirection, groupDescriptor.Member));
                }
            }

            if (sortDescriptors.Any())
            {
                foreach (var sortDescriptor in sortDescriptors)
                {
                    definitions.Add(CreateCustomerSortExpression(sortDescriptor.SortDirection, sortDescriptor.Member));
                }
            }

            return (definitions.Count == 0) ? null : definitions;
        }

        public static IEnumerable<SortSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer>> TranslateustomerSort(IList<GroupDescriptor> groupDescriptors, IList<SortDescriptor> sortDescriptors)
        {
            var definitions = new List<SortSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer>>();

            if (groupDescriptors.Any())
            {
                foreach (var groupDescriptor in groupDescriptors.Reverse())
                {
                    definitions.Add(CreateCustomerSortExpression(groupDescriptor.SortDirection, groupDescriptor.Member));
                }
            }

            if (sortDescriptors.Any())
            {
                foreach (var sortDescriptor in sortDescriptors)
                {
                    definitions.Add(CreateCustomerSortExpression(sortDescriptor.SortDirection, sortDescriptor.Member));
                }
            }

            return (definitions.Count == 0) ? null : definitions;
        }

        /// <summary>
        /// Creates the order sort expression.
        /// </summary>
        /// <param name="sortDirection">The sort direction.</param>
        /// <param name="memberName">Name of the member.</param>
        /// <returns></returns>
        private static SortSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer> CreateCustomerSortExpression(ListSortDirection sortDirection, string memberName)
        {
            var definition = new SortSpecification<Volvo.CustomerMaster.Customer.Model.Entities.Customer> { Direction = sortDirection };

            switch (memberName)
            {
                case "Number":
                    definition.Predicate = order => order.Number;
                    break;


                case "Name":
                    definition.Predicate = order => order.Name;
                    break;

                case "Address":
                    definition.Predicate = order => order.Address;
                    break;
            }

            return definition;
        }        
    }
}