﻿using System.Collections.Generic;
using Microsoft.Practices.EnterpriseLibrary.Security;
using Microsoft.Practices.Unity;
using Volvo.CustomerMaster.Infrastructure.Common.Utilities.Entities;
using Volvo.CustomerMaster.Infrastructure.Common.Utilities.Unity;
using Volvo.POS.UserDomain.ServiceLayer;
using Volvo.POS.UserDomain.DomainLayer;
using Volvo.CustomerMaster.Web.Common.Session;

namespace Volvo.CustomerMaster.Web.Common.Security
{
    public class SecurityHelper
    {

        [Dependency]
        public static IUserService UserService { get; set; }

        /// <summary>
        /// Authenticate the user to verify that the user is a logged in user and that the user is approved 
        /// by the external authorization system.
        /// </summary>
        /// <returns></returns>
        public static bool Authenticate()
        {
            // Inject implementation of the UserService through DI
            if (UserService == null)
            {
                UserService = Container.Resolve<UserService>();
            }

            string userName = GetWindowsVcnUserName();

            // Get user from external authorization system
            GenericUser user = UserService.GetUser(userName);
            if (user == null)
            {
                return false;
            }

            // Set session 
            SessionWrapper.CurrentUser = user;
            return true;
        }


        /// <summary>
        /// Returns true if the user contain roles that is valid for selected rule
        /// </summary>
        /// <param name="rule"></param>
        /// <returns></returns>
        public static bool Authorized(string rule)
        {
            try
            {
                IList<string> rules = new List<string> { rule };
                return Authorized(rules);
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Returns true if the user contain roles that is valid for selected rules
        /// </summary>
        /// <param name="rules"></param>
        /// <returns></returns>
        public static bool Authorized(IList<string> rules)
        {
            // If user is not defined, try to authenticate it
            if (SessionWrapper.CurrentUser == null)
            {
                if (!Authenticate())
                {
                    return false;
                }
            }

            // Get authorization provider from Entlib
            IAuthorizationProvider auth = AuthorizationFactory.GetAuthorizationProvider("RulesProvider");

            if (rules.Count > 0 && SessionWrapper.CurrentUser.Principal != null)
            {
                foreach (string rule in rules)
                {
                    // Authorize user (with its roles) agains the rule
                    if (!auth.Authorize(SessionWrapper.CurrentUser.Principal, rule))
                    {
                        return false;
                    }
                }
            }
            else
            {
                return false;
            }
            return true;
        }

        private static string GetWindowsVcnUserName()
        {
            // Get windows user
            System.Security.Principal.WindowsIdentity loggedInUser =
                    System.Security.Principal.WindowsIdentity.GetCurrent();

            if (loggedInUser != null)
            {
                string username = loggedInUser.Name;
                username = username.Substring(username.IndexOf('\\') + 1);
                username = username.ToUpper();

                return username;
            }
            return null;
        }
    }
}
