﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Web.Mvc;
using Volvo.CustomerMaster.Web;

namespace Volvo.CustomerMaster.Web.Common.Validation
{
    /// <summary>
    /// This Attribute compares two propertis according with CompareToOperation.
    /// The types of fields must be IComparable.
    /// </summary>
    [AttributeUsage(AttributeTargets.Class | AttributeTargets.Property, AllowMultiple = true, Inherited = false)]
    public class CompareToAttribute : ValidationAttribute, IClientValidatable, IMetadataAware
    {
        public CompareToOperation CompareToOperation { get; set; }

        public string ClientPrefix { get; set; }
        public string ClientSufix { get; set; }
        public string PropertyNaneTarget { get; set; }

        protected override ValidationResult IsValid(object value, ValidationContext validationContext)
        {
            if (value == null)
                return new ValidationResult(string.Format(UI_Resources.Exc_Msg_NullValue, validationContext.DisplayName));

            var entityObject = validationContext.ObjectInstance;
            var entityType = entityObject.GetType();

            var propertyObject = value;
            var propertyType = value.GetType();

            // Verifying if the properties exist
            if (entityType.GetProperties().Count(query => query.Name == PropertyNaneTarget) != 1)
            {
                throw new Exception("Properties don't exist on target class type.");
            }

            var targetValue = entityType.GetProperties().First(query => query.Name == PropertyNaneTarget).GetValue(entityObject, null);

            if (targetValue.GetType() != propertyType)
            {
                throw new Exception("The types of values of properties are different.");
            }

            var comp1 = propertyObject as IComparable;
            var comp2 = targetValue as IComparable;

            if (comp1 == null || comp2 == null)
                throw new Exception("The properties are not IComparable.");

            var result = comp1.CompareTo(comp2);

            var isValid = false;

            switch (CompareToOperation)
            {
                case CompareToOperation.LessThan:
                    isValid = result == -1;
                    break;

                case CompareToOperation.EqualTo:
                    isValid = result == 0;
                    break;

                case CompareToOperation.GreaterThan:
                    isValid = result == 1;
                    break;

                default:
                    isValid = true;
                    break;
            }

            return isValid
                       ? ValidationResult.Success
                       : new ValidationResult(string.Format(UI_Resources.Exc_Msg_CompareToValidator, propertyType.Name, CompareToOperation, PropertyNaneTarget));
        }
        
        public IEnumerable<ModelClientValidationRule> GetClientValidationRules(ModelMetadata metadata, ControllerContext context)
        {
            // translate to html id
            var targetPropertyClientId = string.Format("{0}_{1}{2}", ClientPrefix, PropertyNaneTarget, ClientSufix);
            var currentPropertyClientId = string.Format("{0}_{1}{2}", ClientPrefix, metadata.PropertyName, ClientSufix);
            
            return new[]
                       {
                           new ModelClientValidationCompareToRule(FormatErrorMessage(metadata.GetDisplayName()), currentPropertyClientId, targetPropertyClientId, CompareToOperation)
                       };
        }

        public void OnMetadataCreated(ModelMetadata metadata)
        {
            metadata.DataTypeName = "String";
        }
    }
    
    public enum CompareToOperation
    {
        EqualTo = 0,
        LessThan = 1,
        GreaterThan = 2
    }

}
