﻿/**
* This file contains some base behaviors examples using jquery
*/
(function ($) {
    /**
    * The $.NVS object contains all utilities functions regarding NVS solutions
    */
    if ($.NVS == undefined)
        $.NVS = {};

    $.extend($.NVS, {
        regexNumber: /^([+-]?(((\d+(\.)?)|(\d*\.\d+))([eE][+-]?\d+)?))$/,

        /// <summary>
        ///     Disables/enables NVS specific extensions for telerik controls
        /// </summary>
        loadTelerikExtensions: true,

        /// <summary>
        /// Set if the grid will avoid any user actions until it is completelly loaded.
        /// </summary>
        isGridSafeLoading: true,

        /// <summary>
        ///    showSummaryType:
        ///       inline - default
        ///       alert
        /// </summary>
        showSummaryType: "inline",

        alertErrorMessages: function (messages) {
            /// <summary>
            ///     Show an alert with all messages.
            /// </summary>
            /// <param name="messages" type="Array">
            ///     An array of strings.
            /// </param>
            
            var mountedMessages = new String();

            for (var i = 0; i < messages.length; i++) {
                mountedMessages += '\n' + messages[i];
            }

            alert(mountedMessages.substring(1, mountedMessages.length));
        },

        appendInlineErrorMessages: function (messages) {
            /// <summary>
            ///     Append a set of inline messages on summary.
            /// </summary>
            /// <param name="messages" type="array">
            ///     An array of messages.
            /// </param>

            var container = $("div[data-valmsg-summary=true]"), list = container.find("ul");

            if (list && list.length && messages && messages.length) {
                list.empty();
                container.addClass("validation-summary-errors").removeClass("validation-summary-valid");

                for (var i = 0; i < messages.length; i++) {
                    $("<li />").html(messages[i]).appendTo(list);
                }
            }
        },

        appendInlineInfoMessage: function (message) {
            /// <summary>
            ///     Append an inline message on summary.
            /// </summary>
            /// <param name="message" type="String">
            ///     A message.
            /// </param>

            if ($('.messageContainer .message.info').length == 0) {
                $('.messageContainer').prepend('<div id="messages" class="message info"><ul class="info"></ul></div>');
            }

            $("<li />").html(message).appendTo($(".messageContainer .info ul"));

            // Fixed the scroll issue with jQuery layout plugin
            var scrollTopOffset = $('.ui-layout-center .panel').scrollTop();
            $.NVS.panelSetLayout.sizeContent('center');
            $('.ui-layout-center .panel').scrollTop(scrollTopOffset);
        },

        clearErrorSummary: function () {
            /// <summary>
            ///     Clear the error summary.
            /// </summary>
            
            var $summary = $('div[data-valmsg-summary=true]');
            $summary.removeClass('validation-summary-errors');
            $summary.addClass('validation-summary-valid');
            $('ul li', $summary).remove();
            $('ul', $summary).append('<li style="display: none;"></li>');

            $('#validationFlashMessages .info').remove();
        },

        collapsible: function ($element) {
            /// <summary>
            ///     Collapsible the panel.
            /// </summary>
            /// <param name="$element" type="jQuery Object">
            ///     Collapse the panel element.
            /// </param>

            if ($element.hasClass("closed")) {
                $element.find("div").addClass("closed").addClass("collapsibleContainerIsClosed");
            }
            $element.find("h3:first").bind("click", function (evt) {
                $item = $(this).parent();
                $item.toggleClass("closed");
                if ($item.find("div").hasClass("closed")) {
                    $item.find("div").toggleClass("closed").show(0);
                } else {
                    $item.find("div").toggleClass("closed").hide(0);
                }
            });
        },

        createLoadingBox: function ($grid) {
            /// <summary>
            ///     Creates the gray box in front of Grid for loading process.
            /// </summary>
            /// <param name="$grid" type="jQuery Object">
            ///     The grid element.
            /// </param>
            if (this.isGridSafeLoading) {
                $("<div class='t-grid-loading t-overlay'/>")
                    .css({ top: $grid.scrollTop(), left: $grid.scrollLeft(), width: $grid.width(), height: $grid.height() })
                    .appendTo($grid);
            }
        },

        displayErrors: function (modelState) {
            /// <summary>
            ///     Display the errors from ModelState.
            /// </summary>
            /// <param name="modelState" type="object">
            ///     An object of model state.
            /// </param>
            
            var errorList = new Array();

            for (var prop in modelState) {
                if (modelState[prop] && modelState[prop].errors) {
                    for (var i = 0; i < modelState[prop].errors.length; i++) {
                        errorList.push(modelState[prop].errors[i]);
                    }
                }
            }

            this.displayErrorsFromList(errorList);
        },

        displayErrorsFromList: function (errorList) {
            /// <summary>
            ///     Display the errors from a array.
            /// </summary>
            /// <param name="errorList" type="Array">
            ///     An array of errors (string).
            /// </param>
            switch (this.showSummaryType) {
                case "inline":
                    this.appendInlineErrorMessages(errorList);
                    break;

                case "alert":
                    this.alertErrorMessages(errorList);
                    break;

                default:
                    throw new Error("Undefined showSummaryType");
            }
        },

        displayErrorsFromValidatorErrorList: function (validatorErrorList) {
            /// <summary>
            ///     Display the errors from a validator errorlist.
            /// </summary>
            /// <param name="validatorErrorList" type="Object">
            ///     An array of validator errors.
            /// </param>
            var errorList = new Array();

            for (var i = 0; i < validatorErrorList.length; i++) {
                errorList.push(validatorErrorList[i].message);
            }

            this.displayErrorsFromList(errorList);
        },

        // Calcultes the week number of a specific date
        getWeekNumber: function (date) {
            /// <summary>
            ///     Get the week number of a date.
            /// </summary>
            /// <param name="date" type="Date">
            ///     A date.
            /// </param>
            
            if (!(date instanceof Date))
                return null;

            var year = date.getFullYear();
            var newYear = new Date(year, 0, 1);
            var y2k = (year < 1000) ? year + 1900 : year;

            // uses monday as the first day on a week
            var dayOfWeek = (date.getDay() < 1) ? 6 : date.getDay() - 1;

            // according of ISO 8601, the first week stars until thursday
            var offset = (dayOfWeek < 4) ? 0 : 1;
            var daynum = ((Date.UTC(y2k, date.getMonth(), date.getDate(), 0, 0, 0) - Date.UTC(y2k, 0, 1, 0, 0, 0)) / 1000 / 60 / 60 / 24) + 1;
            var weeknum = Math.floor(daynum / 7) + 1 - offset;

            if (weeknum == 0) {
                var prevNewYear = new Date(year - 1, 0, 1);
                dayOfWeek = (prevNewYear.getDay() < 1) ? 6 : prevNewYear.getDay() - 1
                weeknum = (dayOfWeek < 4) ? 53 : 52;
            }
            else if (weeknum == 52 || weeknum == 53) {
                var nextNewYear = new Date(year + 1, 0, 1);
                dayOfWeek = (nextNewYear.getDay() < 1) ? 6 : nextNewYear.getDay() - 1
                weeknum = (dayOfWeek < 4) ? 1 : weeknum;
            }
            return weeknum;
        },

        getDataType: function (content) {
            /// <summary>
            ///     Identify the type of the content of a string.
            ///     The return value means:
            ///     0 - Undefined
            ///     1 - Number (int, or float, or double, etc)
            ///     2 - Date (using telerik)
            ///     3 - String
            /// </summary>
            /// <param name="data" type="string">
            ///     A string.
            /// </param>

            if (this.regexNumber.test(content))
                return 1;

            if ($.telerik && $.telerik.datepicker.prototype.parse(content, $.telerik.cultureInfo.shortDate) != null) {
                return 2;
            }

            if (typeof (content) == 'string')
                return 3;

            return 0;
        },

        go: function (url) {
            /// <summary>
            ///     Go to another URL.
            /// </summary>
            /// <param name="url" type="string">
            ///     An url.
            /// </param>
            
            window.location = url;
        },

        isClickEventAttached: function (element) {
            /// <summary>
            ///     Checks if the event click is already attached on a spefic element.
            /// </summary>
            /// <param name="element" type="html object">
            ///     An html element.
            /// </param>
            
            var events = $.data(element, 'events');
            return events != undefined && events.click != undefined;
        },

        localization: {
            TelerikGridSavedRow: 'The ordered Part was saved.',
            TelerikGridSavedAllRows: 'All ordered Parts were saved.',
            TelerikGridAddedRow: 'The ordered Part was added.',
            TelerikGridDeletedRow: 'The ordered Part was deleted.',
            TelerikGridToolbarItemEditAllRows: 'Edit All Rows',
            TelerikGridToolbarItemSaveAllRows: 'Save All Rows',
            TelerikGridPageNumber: 'Go To',
            TelerikSaveCurrentAction: 'Do you want to save the changes?'
        },

        removeLoadingBox: function ($grid) {
            /// <summary>
            ///     Remove the loading box.
            /// </summary>
            /// <param name="$grid" type="jQuery Object">
            ///     The grid element.
            /// </param>
            if (this.isGridSafeLoading) {
                $grid.find(".t-grid-loading").remove();
            }
        },

        sumChildrenUsedHeight: function ($element) {
            /// <summary>
            ///     Sum the Children used height.
            /// </summary>
            /// <param name="$element" type="jQuert object">
            ///     An jQuery element.
            /// </param>
            
            var realContentHeight = 0;
            var $children = $element.children();
            for (var i = 0; i < $children.length; i++) {
                realContentHeight += $($children[i]).outerHeight(true);
            }
            return realContentHeight;
        },

        updateGridsHeight: function () {
            /// <summary>
            ///     Update all Grid's Height
            /// </summary>

            $('.t-grid').each(function () {
                var $grid = $(this);
                var telerikGrid = $grid.data('tGrid');
                if (telerikGrid.autoExpandable) {
                    var $parentDiv = $grid.parents(':first');
                    var contentHeight = $.NVS.sumChildrenUsedHeight($parentDiv);
                    var offsetHeight = $parentDiv.innerHeight() - contentHeight;
                    var $gridContent = $('.t-grid-content', $grid);
                    $gridContent.height($gridContent.height() + offsetHeight);
                }
            });
        }
    });
    $(window).ready(function () {
        //Set up  jQuery layout plugin to resize main footer and content area properly
        var panelSetOptions = {
            defaults: {
                contentSelector: ".panel",
                spacing_open: 0,
                spacing_closed: 0,
                findNestedContent: true,
                showOverflowOnHover: false,
                applyDefaultStyles: true
            },
            west: { initHidden: true },
            east: { initHidden: true },
            north: { applyDefaultStyles: false },
            south: { applyDefaultStyles: false }

        };
        var panelSetLayout = $('#panelSet').layout(panelSetOptions);

        $.extend($.NVS, {
            panelSetLayout: panelSetLayout
        });

        // Register all collapsible groups behavior
        $(".collapsible").each(function () { $.NVS.collapsible($(this)); });

        /******************** Validation  ******************/
        // Code to change the place where will apear the validation summary from MVC3 framework
        var $form = $('form div.validation-summary-errors[data-valmsg-summary=true]');
        if ($form.length > 0) {
            var $existentValidationBox = $('#validationFlashMessages .message');
            $('#validationSummary').remove();
            if ($existentValidationBox.length > 0) {
                $existentValidationBox.removeClass('message');
                $existentValidationBox.addClass('validation-summary-errors');

                if ($('ul.error', $existentValidationBox).length == 0)
                    $($existentValidationBox).append('<ul class="error"></ul>');

                $('ul.error', $existentValidationBox).append($('ul li', $form));
                $form.remove()
            }
            else {
                $('#validationFlashMessages').append($form);
            }
        }
        else {
            $form = $('form div[data-valmsg-summary=true]');
            if ($form.length > 0) {
                $('#validationFlashMessages').append($form);
            }
            // if none form was created with summary, create the summary block
            else {
                $('#validationFlashMessages').append('<div data-valmsg-summary="true" class="validation-summary-valid"><ul><li style="display: none;"></li></ul></div>');
            }
        }
    });

    window.jQuery.NVS = window.$.NVS = $.NVS;
})(jQuery);

